/*

  testfile.c
  
  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (C) 1996,1997 SSH Communications Security Oy, Espoo, Finland
  All rights reserved.

  Created: Wed Jul 23 22:36:43 1997 [mkojo]

  Testing some things with GMP etc. 

  */

/*
 * $Id: t-sophie-germain.c,v 1.2 1998/10/08 08:38:02 mkojo Exp $
 * $Log: t-sophie-germain.c,v $
